from newick.backend.nhx_util import *

def test_generate_nhx_null():
    assert generate_nhx(None) == ""
    
def test_generate_nhx_empty():
    assert generate_nhx(dict()) == ""
    
def test_generate_nhx_alphanum():
    dct = {"A":1, "Bonn": None, 52: True}
    assert generate_nhx(dct) == "[&&NHX:A=1:Bonn=None:52=True]"

def test_generate_nhx_othersym():
    dct = {"A=C":1, "B:nn": "B(er)lin", "new\n-line": "s p a c e"}
    assert generate_nhx(dct) == "[&&NHX:A\\=C=1:B\\:nn=B\\(er\\)lin:new\\\n-line=s p a c e]"
    
def test_nhx_filter_str_0():
    assert nhx_filter_str("A=C") == "A\\=C"
    


from newick.backend.tree import Tree
from newick.backend.node import RootNode, Node
from newick.backend.path import Path
from newick.backend.util_funcs import format_int
from typing import Callable
from enum import Enum


class BlacklistTokenStrat(Enum):
    IGNORE_BLACKLIST = 0
    DROP_TOKEN = 1
    DROP_AFTER_FIRST = 2
    DROP_ENTIRE_LINE = 3
    

def tree_parse_basic(text:str, 
                     root_label:str=None, 
                     line_delim:str=";", 
                     waypoint_sep:str=",",
                     label_dist_sep:str=":", 
                     trim_sym:str='\r\n ',
                     blacklist:list[str]=["n.a.", "O", "Unclassified"],
                     blacklist_token_strat:BlacklistTokenStrat=BlacklistTokenStrat.DROP_AFTER_FIRST,
                     default_dist:float=1.0,
                     dist_adjust_strategy:Callable[[Node,float],float]=None) -> Tree:
    index = 0
    lines = text.split(line_delim)
    outtree = Tree(RootNode(root_label), 
                   default_dist=default_dist)
    outtree.set_dist_adjust_strat(dist_adjust_strategy)
    for line in lines:
        line = clean_token(line, trim_sym)
        if line != "":
            waypoints = line.split(waypoint_sep)
            outpath = Path(root_label=root_label)
            for waypoint in waypoints:
                flag_drop_after_token = False
                waypoint = clean_token(waypoint, trim_sym)
                waypoint_split = waypoint.split(label_dist_sep)
                if (waypoint in blacklist \
                        or waypoint_split[0] in blacklist):
                    match blacklist_token_strat:
                        case BlacklistTokenStrat.DROP_ENTIRE_LINE:
                            outpath = None
                            break
                        case BlacklistTokenStrat.DROP_TOKEN:
                            break
                        case BlacklistTokenStrat.DROP_AFTER_FIRST:
                            flag_drop_after_token = True
                if len(waypoint_split) == 1:
                    ndist = float("-inf")
                else:
                    ndist = float(waypoint_split[1])
                nlabel = waypoint_split[0]
                outpath.add(nlabel, ndist)
                if flag_drop_after_token:
                    break
            if outpath:
                myaddinfo = {"_parse_index": { format_int(index) }}
                outtree.add_new_node(outpath, 
                                    additional_info=myaddinfo)
        index += 1
    return outtree

def clean_token(token, trim_sym):
    return token.strip(trim_sym)


from newick.backend.tree import Tree
from newick.backend.path import Path


def test_basic():
    r = Tree.RootNode("R")
    t = Tree(r)
    t.add_new_node(Path("R", [("A", 1.0)]))
    t.add_new_node(Path("R", [("B", 2.0)]))
    assert t.to_string() == "(A:1,B:2)R;"

def test_partial_duplicate():
    r = Tree.RootNode("R")
    t = Tree(r)
    t.add_new_node(Path("R", [("A", 1.0)]))
    t.add_new_node(Path("R", [("B", 4.0)]))
    t.add_new_node(Path("R", [("B", 2.0), ("C-c", 2.2)]))
    assert t.to_string() == "(A:1,(C-c:2.2)B:3)R;" # note: dist of B changed
    assert t._root.get_child_by_label("B").get_duplication_count() == 0
    t.add_new_node(Path("R", [("B", 4.0)]))
    assert t.to_string() == "(A:1,(C-c:2.2)B:3.5)R;" # note: changed again
    assert t._root.get_child_by_label("B").get_duplication_count() == 1
    
def test_partial_duplicate_no_dist_def():
    r = Tree.RootNode("R")
    t = Tree(r)
    t.add_new_node(Path("R", [("A", 1.0)]))
    t.add_new_node(Path("R", [("B", 4.0)]))
    t.add_new_node(Path("R", [("B", float("-inf")), ("C-c", 2.2)]))
    assert t.to_string() == "(A:1,(C-c:2.2)B:4)R;"
    assert t._root.get_child_by_label("B").get_duplication_count() == 0
    t.add_new_node(Path("R", [("B", float("-inf"))]))
    assert t.to_string() == "(A:1,(C-c:2.2)B:4)R;"
    assert t._root.get_child_by_label("B").get_duplication_count() == 1
    
def test_exact_duplicate():
    r = Tree.RootNode("R")
    t = Tree(r)
    t.add_new_node(Path("R", [("A", 1.0)]))
    t.add_new_node(Path("R", [("B", 2.0), ("C-c", 2.2)]))
    t.add_new_node(Path("R", [("B", 4.0), ("C-c", 2.2)]))
    assert t.to_string() == "(A:1,(C-c:2.2)B:3)R;"
    assert t._root.get_child_by_label("B").get_duplication_count() == 0
    assert t._root.get_child_by_label("B").get_child_by_label("C-c").get_duplication_count() == 1
    
    


from collections.abc import Mapping
from .nhx_util import generate_nhx
from .util_funcs import format_float, format_int


class Node:
    """
    Represents a single node in the tree.
    """
    
    
    _DEFAULT_OUTPUTLABEL_MAPPER:Mapping['Node',str] = \
        lambda n: n.get_label()
    
    
    # class fields
    #_label
    #_distance          = 1.0
    #_children          = []
    #_dupcount          = 0
    #_additional_info   = None
    #_children_by_label = dict()
    
    
    def __init__(self, 
                 label:str, 
                 distance:float         = 1.0, 
                 duplicates_count:int   = 0, 
                 additional_info:dict   = dict(),
                 children:list          = []):
        """Creates a new node with the given information.

        Args:
            label (str): 
                identifier within parent (usually the visible label) 
                of the node. 
                Has to be unique in its parent (see also 
                `to_string()`)! 
            distance (float, optional): 
                distance from the parent node. Defaults to 1.0.
            duplicates_count (int, optional): 
                duplicate counter -- counts how many other nodes 
                there exist with the same label. Defaults to 0.
            additional_info (dict, optional): 
                dictionary with all additional data you want to attach
                to the node (see also `to_string()`). Defaults to 
                dict().
            children (list, optional): 
                list of children nodes. Defaults to [].
                Please avoid using this if possible.
        """
        
        # argument validation
        if label == None:
            label = ""
        elif not isinstance(label, str):
                label = str(label)
        if not isinstance(distance, float):
            distance = float(distance)
        if distance < 0:
            msg = \
                """
                Distance from parent node must be positive.
                """ 
            raise ValueError(distance, msg)
        if not isinstance(additional_info, dict):
            msg = \
                """
                The `additional_info` must be a dictionary to allow 
                NHX serialization for the output. If you want to attach 
                a different kind of object, do so by wrapping it in a 
                dictionary under a static key. 
                """
            raise ValueError(additional_info, msg) 
        
        # write to class members
        self._distance          = distance
        self._label             = label
        self._dupcount          = duplicates_count
        self._additional_info   = additional_info
        # handle children
        self._children          = []
        self._children_by_label = dict() # maps label to index of _children
        for c in children:
            self.add_child(c)
        
    
    def contains_child_with_label(self, label:str) -> bool:
        """Determines wether `self` has a child with the `label`.

        Args:
            label (str): label (identifier within parent) to look for.

        Returns:
            bool: 
                `True` iff a child with that `label` does exist here.
        """
        return label in self._children_by_label.keys()
    
    def count_children(self) -> int:
        """Counts `self`'s children.

        Returns:
            int: The number of children `self` has.
        """
        return len(self._children_by_label)
    
    def is_leaf(self) -> bool:
        """Determines whether this node is a leaf or not.

        Returns:
            bool: Return True iff `self` has no children.
        """ 
        return self.count_children() == 0
    
    def is_root(self) -> bool:
        """
        Determines whether this node is a root node.
        This kind of node is not a root node by definition, so that 
        this function always returns a constant `False`.

        Returns:
            bool: Constant False.
        """
        return False
    
    def get_child_by_label(self, label:str) -> 'Node':
        """Gets the child node with the given label.

        Args:
            label (str): label (identifier within parent) to look for.

        Returns:
            Node: 
                The child node withe the given `label` iff it exists 
                in `self`, otherwise `None`.
        """
        if self.contains_child_with_label(label):
            return self._children[self._children_by_label[label]]
        else:
            return None
    
    def add_child(self, 
                  child:'Node', 
                  dist_adjust_strategy=None,
                  count_duplicate:bool=True) -> tuple[bool,'Node']:
        """
        Adds the `child` node to `self`'s children. 
        
        Duplicate labels are forbidden -- duplicates are handled 
        using the `handle_duplicate()` method.
        Override `handle_duplicate()` for custom behaviour. 

        Args:
            child (Node): 
                node to be added.
            dist_adjust_strategy (Callable[[Node,float],float], optional): 
                A function that takes the pre-existing `Node` in the tree and 
                the distance (`float`) of the supposed new node that is to be inserted, 
                and calculates the new distance that will be written into said
                pre-existing `Node` in the tree. 
                This is supposed to allow the user to control which changes are made on 
                the distance of a node when contradicting information are encountered
                during a new node's insertion. 
                You can define your own function or use one of the 
                `Tree._DIST_ADJUST_STRAT_...`s.
        """
        if not self.contains_child_with_label(child.get_label()):
            self._children_by_label[child.get_label()] = len(self._children)
            self._children.append(child)
            return (True, child)
        else:
            ochild = self.get_child_by_label(child.get_label())
            if dist_adjust_strategy:
                adjusted_dist = dist_adjust_strategy(ochild, 
                                                    child.get_distance())
                ochild.set_distance(adjusted_dist)
            ochild.handle_duplicate(child, count=count_duplicate)
            return (False, ochild)
    
    
    def handle_duplicate(self, other:'Node', count=True):
        """
        Handles the case when a second node with the same `label` as 
        `self`'s was to be added to the parent, which is forbidden. 
        By default this function counts duplicates, copies the 
        additional information and otherwise ignores the  `other` 
        child, which could not be added. That's true sisterly affection. 
        During copying of additional info, whenever a key is already
        present within the dictionary, the new value will be ignored,
        unless both are lists or sets, then they will be merged.
        Override this method for custom behaviour. 
        NOTE: Taking away this functionality may break some 
        `Tree._DIST_ADJUST_STRAT_...`s! It is therefore recommended 
        to still call this function, even when overridden by a 
        subclass.

        Args:
            other (Node): 
                Node that was to be added but was declined because it 
                has the same `label` as self.
            count (bool, optional):
                Whether or not to count this duplication.
        """
        # count duplicates
        if count:
            self._dupcount += 1
        # copy additional information
        s_ao = self.get_additional_info()
        o_ao = other.get_additional_info()
        for k in o_ao.keys():
            v = o_ao[k]
            if k in s_ao.keys():
                if s_ao[k] != v:
                    if isinstance(v, set) and isinstance(s_ao[k], set):
                        s_ao[k] = s_ao[k].union(v)
                    elif isinstance(v, list) and isinstance(s_ao[k], list):
                        s_ao[k].extend(v)
                    else:
                        # This key is already present with a different value. 
                        # TODO: What to do here?
                        pass 
            else:
                # add the key and value.
                s_ao[k] = v
        
        
    def get_label(self) -> str:
        """Retrieves the label (identifier) string of `self`.

        Returns:
            str: label (identifier within its parent) of `self`
        """
        return self._label
    
    def get_distance(self) -> float:
        """Retrieves the distance of `self` from its parent.

        Returns:
            float: distance of `self` from its parent
        """
        return self._distance
    
    def set_distance(self, distance):
        """
        Sets the distanec of `self` to the given `val`.
        """
        if not isinstance(distance, float):
            distance = float(distance)
        if distance < 0:
            msg = \
                """
                Distance from parent node must be positive.
                """ 
            raise ValueError(distance, msg)
        self._distance = distance

    def get_duplication_count(self) -> int:
        """Retrieves the duplicate counter's value.

        Returns:
            int: 
                number of duplicates counted by default 
                `handle_duplicate()`
        """
        return self._dupcount
        
    def get_additional_info(self) -> dict:
        """
        Retrieves the additional info dictionary attached to 
        `self`. 

        Returns:
            dict: additional info dictionary attached to `self`
        """
        return self._additional_info
    
    
    def gen_children_strings(self,
                             with_labels:bool, 
                             with_distances:bool, 
                             with_additional_info_nhx:bool, 
                             outputlabel_mapper:Mapping['Node',str]) -> list:
        """generate the strings for all the children.

        Args:
            with_labels (bool): see `to_string()`.
            with_distances (bool): see `to_string()`.
            with_additional_info_nhx (bool): see `to_string()`.
            outputlabel_mapper (bool): see `to_string()`.

        Returns:
            list: of all the children's string representations.
        """
        ret_ch = []
        for child in self._children:
            ret_ch.append(
                child.to_string(with_labels, 
                                with_distances, 
                                with_additional_info_nhx, 
                                outputlabel_mapper))
        return ret_ch
    
    def to_string(self,
                  with_labels:bool=True,
                  with_distances:bool=True,
                  with_additional_info_nhx:bool=False,
                  outputlabel_mapper:Mapping['Node',str]=None) -> str:
        """
        Generates a string representation of `self` and its 
        children in newick format.

        Args:
            with_labels (bool, optional): 
                Whether or not to output labels (see also 
                `outputlabel_mapper`). Defaults to True.
            with_distances (bool, optional): 
                Whether or not to output distances to parent node. 
                Defaults to True.
            with_additional_info_nhx (bool, optional): 
                Whether to output the additional info attached 
                to this node (in NHX format) or to leave it out. 
                Defaults to False.
                To ensure that the output can be used by third party
                implementations, make sure the string representation
                of each element (key or value) in the dictionary is 
                free from special symbols and control characters
                such as newline, '=' and ':'. 
                Best practice: Only use alphanumerics and pre-convert
                values (and keys) to strings. All other characters 
                will be escaped, hoping for compliance with your 
                other tool's definitions of the syntax of NHX.
                Defaults to False.
            outputlabel_mapper (Mapping[Node,str], optional): 
                A mapping function (`Node` to `str`) that maps a node
                to the label string that is to be written for it in 
                the output. 
                Defaults to `lambda n: n.get_label()`, so that a 
                `Node` is mapped to its `Node._label`, which is the 
                identification `label` (unique in its parent). 

        Returns:
            str: A string representation of `self` and its subtree.
        """
        ret = []
        # append children info
        ret_ch = self.gen_children_strings(with_labels, 
                                           with_distances, 
                                           with_additional_info_nhx, 
                                           outputlabel_mapper)
        if len(ret_ch) > 0:
            ret.append('(' + ','.join(ret_ch) + ')')
        # append own info
        if with_labels:
            if outputlabel_mapper:
                ret.append(outputlabel_mapper(self))
            else:
                ret.append(self._DEFAULT_OUTPUTLABEL_MAPPER())
        if with_additional_info_nhx:
            ret.append(generate_nhx(self.get_additional_info()))
        if with_distances:
            ret.append(':' + format_float(self.get_distance()))
        # convert to string and return
        return ''.join(ret)
    
    
    def __repr__(self) -> str:
        """`to_string()` with default settings.

        Returns:
            str: 
                A string representation of `self` and its subtree with
                default settings.
        """
        return self.to_string()
    
    
class HybridNode(Node):
    """
    Represents a single hybrid node in the tree 
    (see Extended Newick format).
    """
    
    
    # class fields:
    #_hybrid_ignore_pool
    #_hybrid_id
    
    
    _DEFAULT_OUTPUTLABEL_MAPPER:Mapping['Node',str] = \
        lambda n: (n.get_label() + n.gen_hybrid_id_string())
        
    
    def __init__(self, 
                 label:str, 
                 hybrid_id:int,
                 hybrid_ignore_pool:set,
                 distance:float         = 1.0, 
                 duplicates_count:int   = 0, 
                 additional_info:dict   = dict(),
                 children:list          = []):
        """Creates a new node with the given information.

        Args:
            label (str): 
                identifier within parent (usually the visible label) 
                of the node. 
                Has to be unique in its parent (see also 
                `to_string()`)! 
            distance (float, optional): 
                distance from the parent node. Defaults to 1.0.
            hybrid_id (int):
                the identifier number of this hybrid.
            hybrid_ignore_pool (set):
                This mechanism is supposed to prevent duplicate 
                declarations on hybrids. 
                If the hybrid node finds a representation of itself in
                this pool, it will not append its children and 
                additional info to its label during `to_string`.
                This is supposed to be a reference to a set inside the 
                managing `Tree`, which is being flushed by said `Tree` 
                every time a new `to_string` process is being started. 
                When this hybrid node's `to_string` is called, it adds 
                itself to the pool automatically. If you want to 
                prevent that, pass `None` for this arg or override the 
                `handle_to_string()` function.   
            duplicates_count (int, optional): 
                duplicate counter -- counts how many other nodes 
                there exist with the same label. Defaults to 0.
            additional_info (dict, optional): 
                dictionary with all additional data you want to attach
                to the node (see also `to_string()`). Defaults to 
                dict().
            children (list, optional): 
                list of children nodes. Defaults to [].
                Please avoid using this if possible.
        """
        
        # instantiate via super constructor
        super(HybridNode, self).__init__( \
                         label=label, 
                         distance=distance, 
                         duplicates_count=duplicates_count,
                         additional_info=additional_info,
                         children=children)
        
        # argument validation
        if not isinstance(hybrid_id, int):
            msg = \
                """
                Please pass a valid integer for the `hybrid_id`.
                If you don't wish to instantiate a hybrid, use
                the `Node` 
                """
        if not (isinstance(hybrid_ignore_pool, set) 
                or hybrid_ignore_pool == None):
            msg = \
                """
                The `hybrid_ignore_pool` has to be a `set` or `None`.
                """
            raise ValueError(hybrid_ignore_pool, msg)
        
        # write to class fields
        self._hybrid_id = hybrid_id
        self._hybrid_ignore_pool = hybrid_ignore_pool
    
    def gen_hybrid_id_string(self) -> str:
        """
        Generates the portion of the id that contains `self`'s 
        `hybrid_id`.
        
        Returns:
            str: the portion of the id that contains `self`'s 
            `hybrid_id`
        """
        return "#" + format_int(self._hybrid_id)
    
    def gen_children_strings(self,
                             with_labels, 
                             with_distances, 
                             with_additional_info_nhx, 
                             outputlabel_mapper) -> list:
        """
        Generate the strings for all the children.
        Modified so that it returns an empty list if this hybrid is in 
        the `_hybrid_ignore_pool`.

        Args:
            with_labels (bool): see `to_string()`.
            with_distances (bool): see `to_string()`.
            with_additional_info_nhx (bool): see `to_string()`.
            outputlabel_mapper (bool): see `to_string()`.

        Returns:
            list: of all the children's string representations.
        """
        hid_str = self.gen_hybrid_id_string()
        if hid_str in self._hybrid_ignore_pool:
            return [] # on empty list, children will not be included 
                      # in the output string
        else:
            return super(HybridNode, self) \
                .gen_children_strings(with_labels,
                                      with_distances,
                                      with_additional_info_nhx,
                                      outputlabel_mapper)
                
    
    def to_string(self,
                  with_labels:bool=True,
                  with_distances:bool=True,
                  with_additional_info_nhx:bool=False,
                  outputlabel_mapper:Mapping['Node',str]=None) -> str:
        # only print distance when first appearance
        hid_str = self.gen_hybrid_id_string()
        m_with_distances = with_distances and \
                hid_str in self._hybrid_ignore_pool
        return super(HybridNode, self) \
                    .to_string(with_labels,
                               m_with_distances,
                               with_additional_info_nhx,
                               outputlabel_mapper)


class RootNode(Node):
    """
    Represents the root node of the tree.
    """
    
    
    def __init__(self, 
                 label:str, 
                 additional_info:dict   = dict(),
                 children:list          = []):
        """Creates a new node with the given information.

        Args:
            label (str): 
                identifier within parent (usually the visible label) 
                of the node. 
                Has to be unique in its parent (see also 
                `to_string()`)!
            additional_info (dict, optional): 
                dictionary with all additional data you want to attach
                to the node (see also `to_string()`). Defaults to 
                dict().
            children (list, optional): 
                list of children nodes. Defaults to [].
                Please avoid using this if possible.
        """
        super(RootNode, self).__init__(label=label, 
                                       distance=0,
                                       duplicates_count=0,
                                       additional_info=additional_info,
                                       children=children)
        
        
    def from_node(node:Node) -> 'RootNode':
        """
        Create a `RootNode` from the given `node`.
        Note that not all information will be copied from the `node`.
        The following data is guaranteed to be retained:
         * label id string
         * additional info dictionary
         * children nodes
        Other data in `node` may be lost.

        Args:
            node (Node): original node

        Returns:
            RootNode: a clone of `node` as a `RootNode` 
        """
        ret = RootNode(label=node._label,
                       additional_info=node._additional_info)
        ret._children = node._children
        ret._children_by_label = node._children_by_label
        return ret


    def is_root(self) -> bool:
        """
        Determines whether this node is a root node.
        Nodes of type `RootNode`, like this one, are roots by 
        definition.

        Returns:
            bool: Constant True.
        """
        return True
    

    def to_string(self,
                  with_labels:bool=True,
                  with_distances:bool=True,
                  with_additional_info_nhx:bool=False,
                  outputlabel_mapper:Mapping['Node',str]=None) -> str:
        """
        Generates a string representation of `self` and its 
        children in newick format.

        Args:
            with_labels (bool, optional): 
                Whether or not to output labels (see also 
                `outputlabel_mapper`). Defaults to True.
            with_distances (bool, optional): 
                Whether or not to output distances to parent node. 
                Defaults to True.
            with_additional_info_nhx (bool, optional): 
                Whether to output the additional info attached 
                to this node (in NHX format) or to leave it out. 
                Defaults to False.
                To ensure that the output can be used by third party
                implementations, make sure the string representation
                of each element (key or value) in the dictionary is 
                free from special symbols and control characters
                such as newline, '=' and ':'. 
                Best practice: Only use alphanumerics and pre-convert
                values (and keys) to strings. All other characters 
                will be escaped, hoping for compliance with your 
                other tool's definitions of the syntax of NHX.
                Defaults to False.
            outputlabel_mapper (_type_, optional): 
                A mapping function (`Node` to `str`) that maps a node
                to the label string that is to be written for it in 
                the output. 
                Defaults to `lambda n: n.get_label()`, so that a 
                `Node` is mapped to its `Node._label`, which is the 
                identification `label` (unique in its parent). 

        Returns:
            str: A string representation of `self` and its subtree.
        """
        
        ret = []
        # append children info
        ret_ch = self.gen_children_strings(with_labels, 
                                           with_distances, 
                                           with_additional_info_nhx, 
                                           outputlabel_mapper)
        if len(ret_ch) > 0:
            ret.append('(' + ','.join(ret_ch) + ')')
        # append own info
        if with_labels:
            if outputlabel_mapper:
                ret.append(outputlabel_mapper(self))
            else:
                ret.append(self._DEFAULT_OUTPUTLABEL_MAPPER())
        if with_additional_info_nhx:
            ret.append(generate_nhx(self.get_additional_info()))
        # convert to string and return
        return ''.join(ret)
    


"""
This file defines some helper functions that may be used
by several other modules. 
"""


def format_float(num:float) -> str:
    return format(num, 'f').rstrip('0').rstrip('.')

def format_int(num:int) -> str:
    return format(num, 'd')


from newick.backend.node import Node


# BASIC TESTS


def test_simple_instantiation_and_parenting():
    # label only
    node0 = Node("testa")
    assert node0.get_label() == "testa"
    # label and distance
    node1 = Node("testb", distance=2.25)
    assert node1.get_label() == "testb"
    assert node1.get_distance() == 2.25
    # now some parenting
    node0.add_child(node1)
    assert node0.count_children() == 1
    assert node1.count_children() == 0
    assert node0.get_child_by_label("testb") is node1
    assert node0.get_child_by_label("xyz") == None
    assert node1.get_child_by_label("testa") == None
    
def test_repr_default_single():
    node0 = Node("testa")
    assert str(node0) == "testa:1"
    
def test_repr_default_two():
    node0 = Node("A")
    node1 = Node("B", distance=2)
    node0.add_child(node1)
    assert str(node0) == "(B:2)A:1"
    
def test_repr_default_afew():
    node0 = Node("A")
    node1 = Node("B", distance=2)
    node2 = Node("C", distance=1.2)
    node3 = Node("D", distance=5)
    node4 = Node("E")
    node5 = Node("F", distance=4.2)
    node2.add_child(node3)
    node2.add_child(node4)
    node1.add_child(node2)
    node0.add_child(node5)
    node0.add_child(node1)
    assert str(node0) == "(F:4.2,((D:5,E:1)C:1.2)B:2)A:1"
    
def test_repr_nodist_single():
    node0 = Node("testa")
    assert node0.to_string(with_distances=False) == "testa"
    
def test_repr_nodist_two():
    node0 = Node("A")
    node1 = Node("B", distance=2)
    node0.add_child(node1)
    assert node0.to_string(with_distances=False) == "(B)A"
    
def test_repr_nodist_afew():
    node0 = Node("A")
    node1 = Node("B", distance=2)
    node2 = Node("C", distance=1.2)
    node3 = Node("D", distance=5)
    node4 = Node("E")
    node5 = Node("F", distance=4.2)
    node2.add_child(node3)
    node2.add_child(node4)
    node1.add_child(node2)
    node0.add_child(node5)
    node0.add_child(node1)
    assert node0.to_string(with_distances=False) == "(F,((D,E)C)B)A"
    
def test_repr_default_single():
    node0 = Node("testa")
    assert node0.to_string(with_labels=False) == ":1"
    
def test_repr_default_two():
    node0 = Node("A")
    node1 = Node("B", distance=2)
    node0.add_child(node1)
    assert node0.to_string(with_labels=False) == "(:2):1"
    
def test_repr_default_afew():
    node0 = Node("A")
    node1 = Node("B", distance=2)
    node2 = Node("C", distance=1.2)
    node3 = Node("D", distance=5)
    node4 = Node("E")
    node5 = Node("F", distance=4.2)
    node2.add_child(node3)
    node2.add_child(node4)
    node1.add_child(node2)
    node0.add_child(node5)
    node0.add_child(node1)
    assert node0.to_string(with_labels=False) == "(:4.2,((:5,:1):1.2):2):1"
    
def test_repr_labels_only_on_leaves_afew():
    node0 = Node("A")
    node1 = Node("B", distance=2)
    node2 = Node("C", distance=1.2)
    node3 = Node("A", distance=5)
    node4 = Node("E")
    node5 = Node("F", distance=4.2)
    node2.add_child(node3)
    node2.add_child(node4)
    node1.add_child(node2)
    node0.add_child(node5)
    node0.add_child(node1)
    only_leaves_name_mapper = lambda n:  n.get_label() if n.is_leaf() else ""
    assert node0.to_string(outputlabel_mapper=only_leaves_name_mapper) == "(F:4.2,((A:5,E:1):1.2):2):1"
    
def test_duplicates():
    node0 = Node("A")
    node1 = Node("B", distance=2)
    node2 = Node("B")
    assert node0.add_child(node1)[0]
    assert not node0.add_child(node2)[0]
    assert str(node0) == "(B:2)A:1"
    assert node1.get_duplication_count() == 1
    
    
# TODO: HYBRID NODE TESTS




from newick.frontend.very_basic import tree_parse_basic, BlacklistTokenStrat
import os
import pytest


def test_basic0():
    txt = """
    a0,b0,c0,d0;
    a1,b0,c1,d0;
    a0,b1,c2;
    a0,b0,c4,d1;
    """
    t = tree_parse_basic(txt, "r")
    print("---")
    print(t.to_string())
    print("---")
    assert t.to_string() == "((((d0:1)c0:1,(d1:1)c4:1)b0:1,(c2:1)b1:1)a0:1,(((d0:1)c1:1)b0:1)a1:1)r;"
    
def test_basic0_nodist():
    txt = """
    a0,b0,c0,d0;
    a1,b0,c1,d0;
    a0,b1,c2;
    a0,b0,c4,d1;
    """
    t = tree_parse_basic(txt, "r")
    print("---")
    print(t.to_string(with_distances=False))
    print("---")
    assert t.to_string(with_distances=False) == "((((d0)c0,(d1)c4)b0,(c2)b1)a0,(((d0)c1)b0)a1)r;"
    
def test_basic0_nodist_blacklist_dropafterfirst():
    txt = """
    a0,b0,c0,d0;
    a1,b0,c1,d0;
    a0,b1,c2;
    a0,b0,c4,d1;
    a0,n.a.,n.a.;
    """
    t = tree_parse_basic(txt, "r", blacklist_token_strat=BlacklistTokenStrat.DROP_AFTER_FIRST)
    print("---")
    print(t.to_string(with_distances=False))
    print("---")
    assert t.to_string(with_distances=False) == "((((d0)c0,(d1)c4)b0,(c2)b1,n.a.)a0,(((d0)c1)b0)a1)r;"
    
def test_basic0_nodist_blacklist_ignore():
    txt = """
    a0,b0,c0,d0;
    a1,b0,c1,d0;
    a0,b1,c2;
    a0,b0,c4,d1;
    a0,n.a.,n.a.;
    """
    t = tree_parse_basic(txt, "r", blacklist_token_strat=BlacklistTokenStrat.IGNORE_BLACKLIST)
    print("---")
    print(t.to_string(with_distances=False))
    print("---")
    assert t.to_string(with_distances=False) == "((((d0)c0,(d1)c4)b0,(c2)b1,(n.a.)n.a.)a0,(((d0)c1)b0)a1)r;"
    
def test_basic0_nodist_blacklist_drop():
    txt = """
    a0,b0,c0,d0;
    a1,b0,c1,d0;
    a0,b1,c2;
    a0,b0,c4,d1;
    a0,n.a.,n.a.;
    """
    t = tree_parse_basic(txt, "r", blacklist_token_strat=BlacklistTokenStrat.DROP_TOKEN)
    print("---")
    print(t.to_string(with_distances=False))
    print("---")
    assert t.to_string(with_distances=False) == "((((d0)c0,(d1)c4)b0,(c2)b1)a0,(((d0)c1)b0)a1)r;"
        
def test_basic0_nodist_blacklist_dropline():
    txt = """
    a0,b0,c0,d0;
    a1,b0,c1,d0;
    a0,b1,c2;
    a0,b0,c4,d1;
    a0,n.a.,n.a.;
    """
    t = tree_parse_basic(txt, "r", blacklist_token_strat=BlacklistTokenStrat.DROP_ENTIRE_LINE)
    print("---")
    print(t.to_string(with_distances=False))
    print("---")
    assert t.to_string(with_distances=False) == "((((d0)c0,(d1)c4)b0,(c2)b1)a0,(((d0)c1)b0)a1)r;"
    
#@pytest.mark.xfail()
def test_file0():
    #import pdb; pdb.set_trace()
    with open('./test_file0.txt', 'r') as file:
        txt = file.read()
        t = tree_parse_basic(txt, "r")
        with open("./test_file0_out.tree", "w") as out:
            out_str = t.to_string()
            out.write(out_str)
            


def generate_nhx(dict:dict, ext_head='&&NHX') -> str:
    if dict == None or len(dict) == 0:
        return ""
    else:
        ret_elements = []
        for key in dict.keys():
            str_key = nhx_filter_str(str(key))
            str_val = nhx_filter_str(str(dict[key]))
            str_el = str_key + '=' + str_val
            ret_elements.append(str_el)
        str_elements = ':'.join(ret_elements)
        return "[" + ext_head + ':' + str_elements + "]"
    
def nhx_filter_str(string:str):
    global _NHX_FILTER_NO_ESCAPE
    if '_NHX_FILTER_NO_ESCAPE' in globals():
        sym_no_escape = _NHX_FILTER_NO_ESCAPE
    else:
        sym_no_escape =  {'_', '-', '.', '+', '?', '!', '#', '~', 
                        '\'', '%', '§', '$', '€', '&', '/', '@', 
                        '*', ' '}
    skipset = set()
    mod_string = string
    for c in set(string):
        if          not c.isalnum() \
                and not c in sym_no_escape \
                and not c in skipset:
            mod_string = mod_string.replace(c, '\\'+c)
            skipset.add(c)
    return mod_string



from .util_funcs import format_float


class Path:
    
    # class fields
    #_waypoints:list = list()
    #_distances:list = list()
    
    
    def __init__(self, 
                 root_label=None,
                 points:list[tuple[str,float]]=[]):
        self._waypoints = list()
        self._distances = list()
        if root_label:
            self._waypoints.append(root_label)
            self._distances.append(0.0)
        else:
            self._waypoints.append("")
            self._distances.append(0.0)
        for w, d in points:
            self.add(w, d)
    
    
    def __len__(self):
        return len(self._waypoints)
    
    def __getitem__(self, key):
        return [(self._waypoints[i], self._distances[i]) for i in range(len(self))][key]
    
    def __contains__(self, item):
        if      isinstance(item, str):
            return item in self._waypoints
        elif    isinstance(item, int) \
             or isinstance(item, float):
            return item in self._waypoints
        elif    isinstance(item, tuple[str,int]) \
             or isinstance(item, tuple[str,float]):
            return item in zip(self._waypoints, self._distances)
        else:
            raise TypeError(item, "unexpected item type")
    
    def __iter__(self):
        yield from [(self._waypoints[i], self._distances[i]) for i in range(len(self))]
        
        
    def __repr__(self):
        ret_strs = []
        for w,d in self:
            ret_strs.append(w + ":" + format_float(d))
        return type(self).__name__ + "(" + " -> ".join(ret_strs) + ")"
    
    
    def add(self,
            waypoint:str,
            distance:float=float('-inf')):
        """Adds a waypoint to the path.

        Args:
            waypoint (str): 
                label of the waypoint.
            distance (float, optional): 
                distance of this waypoint from the prior.
                Pass `float('-inf')`, which will be replaced by the
                `Tree`'s `default_distance`. 
                Defaults to `float('-inf')`.
        """
        self._waypoints.append(str(waypoint))
        self._distances.append(float(distance))
    
    


from collections.abc import Mapping
from typing import Callable, Iterable
from os import linesep
from .node import Node, HybridNode, RootNode
from .path import Path

class Tree:
    """
    Wraps around the root node of the tree in order to 
    represent the entire tree. Also provides some additional
    functionality.
    """
    
    
    # class fields
    #_root
    #_hybrids  # dict[(str, int), HybridNode] # TODO:
    #_hybrid_ignore_set
    #_default_dist
    #_dist_adjust_strat
  
  
    def _DIST_ADJUST_STRAT_NEW_fn(old_node:Node, new_dist:float) -> float:
        return new_dist
    _DIST_ADJUST_STRAT_NEW=_DIST_ADJUST_STRAT_NEW_fn
    
    def _DIST_ADJUST_STRAT_OLD_fn(old_node:Node, new_dist:float) -> float:
        return new_dist
    _DIST_ADJUST_STRAT_OLD = _DIST_ADJUST_STRAT_OLD_fn
    
    def _DIST_ADJUST_STRAT_ROLL2_fn(old_node:Node, new_dist:float) -> float:
        return (old_node.get_distance() + new_dist) / 2
        
    def _DIST_ADJUST_STRAT_AVERAGE_fn(old_node:Node, new_dist:float) -> float:
        weighted_old_dist = (old_node.get_distance() * (old_node.get_duplication_count()+1))
        return (weighted_old_dist + new_dist) / (old_node.get_duplication_count()+2)
    _DIST_ADJUST_STRAT_AVERAGE=_DIST_ADJUST_STRAT_AVERAGE_fn
    
    
    def __init__(self,
                 root_node:RootNode,
                 default_dist:float=1.0,
                 dist_adjust_strategy:Callable[[Node,float],float]=_DIST_ADJUST_STRAT_AVERAGE):
        """Creates a new node with the given information.

        Args:
            root_node (RootNode): 
                Root node of the tree. 
                Use the `RootNode()` constructor to create one.
            default_dist (float, optional):
                The default distance of a node from its parent in this tree,
                which is used when a path does not contain distance information.
                Defaults to 1.
            dist_adjust_strategy (Callable[[Node,float],float], optional):
                A function that takes the pre-existing `Node` in the tree and 
                the distance (`float`) of the supposed new node that is to be inserted, 
                and calculates the new distance that will be written into said
                pre-existing `Node` in the tree. 
                This is supposed to allow the user to control which changes are made on 
                the distance of a node when contradicting information are encountered
                during a new node's insertion. 
                You can define your own function or use one of the 
                `Tree._DIST_ADJUST_STRAT_...`s.
        """
        
        # validation
        if not isinstance(root_node, RootNode):
            if isinstance(root_node, Node):
                root_node = RootNode.from_node(root_node)
            else:
                msg = \
                """
                The `root_node` has to be a RootNode.
                Use the `RootNode()` constructor to create it.
                """
                raise ValueError(root_node, msg)
        #TODO: validate other args
        
        # write class fields 
        self._root = root_node
        self._default_dist = default_dist
        self._hybrids = dict()
        self._hybrid_ignore_set = set()
        self.set_dist_adjust_strat(dist_adjust_strategy)

    
    def node_dist_or_def(self, 
                         dist:float):
        if dist == float("-inf"):
            return self._default_dist
        else:
            return dist
        
    def dist_adjust_strat_or_def(self, 
                                 dist:float):
        """
        Returns the distance adjustment function only if the given 
        distance is not default (`float("-inf")`).
        This is to prevent distance adjustments after first definition
        if no explicit distances are given. 
        """
        if dist == float("-inf"):
            return None
        else:
            return self._dist_adjust_strat
    
    
    RootNode = RootNode
    
    
    def reg_hybrid_id(self,
                      label:str,
                      hybrid_id:int,
                      distance:float,
                      additional_info:dict) -> HybridNode:
        """For internal use only.

        Returns:
            HybridNode: 
                The HybridNode instance with the given label and id
                that is already present in the tree, or a newly created
                one if it is not registered yet.
        """
        if (label, hybrid_id) in self._hybrids.keys:
            return self._hybrids[(label, hybrid_id)]
        else:
            wchild = HybridNode(label,
                                hybrid_id,
                                distance, 
                                additional_info)
            self._hybrids[(label, hybrid_id)] = wchild
            return wchild
     
    
    def add_new_node(self, path:Path, additional_info:dict=dict()) -> bool:
        """
        Adds a new node to the tree if it does not exist yet in the 
        location determined by the `path`.
        Also attaches the given additional info.
        If the node is already present, by default
         * duplication is counted
         * distance is adjusted by the tree's strategy, if a distance
           is given
         * additional info is being copied over (lists and sets are merged)

        Args:
            path (Path): 
                Path where to place the node.
            additional_info (dict, optional): 
                Additional info dict to attach

        Raises:
            ValueError: When the given path is too short or the root does 
            not match.

        Returns:
            bool: 
                True iff the node has been created, False if it had to be 
                merged.
        """
        cparent = self._root
        cret = False
        # check root
        if (len(path) <= 1):
            msg = \
                "Cannot insert a node with a path shorter than 2 waypoints."
            raise ValueError(path, msg)
        if (self._root.get_label(), self._root.get_distance()) != path[0]:
            msg = \
                "The start waypoint of the path differs from the tree's root."
            raise ValueError(path, msg)
        # insert rest
        cpath = path[1:]
        while len(cpath) > 0:
            is_end_of_path = (len(cpath) == 1)
            wlabel, wdist = cpath[0]
            w_dist_adjust_strat = self.dist_adjust_strat_or_def(wdist)
            wdist = self.node_dist_or_def(wdist)    
            waddinfo = additional_info if is_end_of_path else dict()
            wchild = Node(wlabel, 
                          distance=wdist, 
                          additional_info=waddinfo)
            cret, achild = cparent.add_child(wchild, 
                                     w_dist_adjust_strat,
                                     count_duplicate=is_end_of_path)
            cpath = cpath[1:]
            cparent = achild
        return cret
    
    def add_new_hybrid_node(self, paths:Iterable[str], hybrid_id=-1, additional_info:dict=dict()) -> bool:
        """
        Adds a new hybrid node to the tree if it does not exist yet in the 
        locations determined by the `path`.
        Also attaches the given additional info.
        If the node is already present, by default
         * duplication is counted
         * distance is adjusted by the tree's strategy, if a distance
           is given
         * additional info is being copied over (lists and sets are merged)

        Args:
            path (Iterable[Path]): 
                Collection of paths where to place the node.
            additional_info (dict, optional): 
                Additional info dict to attach

        Raises:
            ValueError: When the given path is too short or the root does 
            not match.

        Returns:
            bool: 
                True iff the node was inserted in at least one location, 
                otherwise False.
        """
        ret = False
        is_first_path = True
        for path in paths:
            cparent = self._root
            cret = False
            # check root
            if (len(path) <= 1):
                msg = \
                    "Cannot insert a node with a path shorter than 2 waypoints."
                raise ValueError(path, msg)
            if (self._root.get_label(), self._root.get_distance()) != path[0]:
                msg = \
                    "The start waypoint of the path differs from the tree's root."
                raise ValueError(path, msg)
            # insert rest
            cpath = path[1:]
            while len(cpath) > 0:
                is_end_of_path = (len(cpath) == 1)
                wlabel, wdist = cpath[0]
                w_dist_adjust_strat = self.dist_adjust_strat_or_def(wdist)
                wdist = self.node_dist_or_def(wdist)    
                if is_end_of_path and is_first_path:
                    waddinfo = additional_info
                else:
                    waddinfo = dict()
                wchild = self.reg_hybrid_id(wlabel, 
                                            hybrid_id,
                                            distance=wdist, 
                                            additional_info=waddinfo)
                cret, achild = cparent.add_child(wchild, 
                                         w_dist_adjust_strat,
                                         count_duplicate=is_end_of_path)
                cpath = cpath[1:]
                cparent = achild
            ret |= cret
            is_first_path = False
        return ret
    
    
    def set_dist_adjust_strat(self, dist_adjust_strat:Callable[[Node,float],float]):
        """Sets the distance adjustment function of this tree. 

        Args:
            dist_adjust_strat (Callable[[Node,float],float])
        """
        if dist_adjust_strat:
            self._dist_adjust_strat = dist_adjust_strat
        else:
            pass  # maybe throw error?
    
    
    def to_string(self,
                  with_labels:bool=True,
                  with_distances:bool=True,
                  with_additional_info_nhx:bool=False,
                  append_newline:bool=False,
                  outputlabel_mapper:Mapping[Node,str]=None) -> str:
        """
        Generates a string representation of this tree in newick 
        format.

        Args:
            with_labels (bool, optional): 
                Whether or not to output labels (see also 
                `outputlabel_mapper`). Defaults to True.
            with_distances (bool, optional): 
                Whether or not to output distances to parent node. 
                Defaults to True.
            with_additional_info_nhx (bool, optional): 
                Whether to output the additional info attached 
                to this node (in NHX format) or to leave it out. 
                Defaults to False.
                To ensure that the output can be used by third party
                implementations, make sure the string representation
                of each element (key or value) in the dictionary is 
                free from special symbols and control characters
                such as newline, '=' and ':'. 
                Best practice: Only use alphanumerics and pre-convert
                values (and keys) to strings. All other characters 
                will be escaped, hoping for compliance with your 
                other tool's definitions of the syntax of NHX.
                Defaults to False.
            default_distance (int, optional):
                The default distance between a node and its parent
                Defaults to 1.
            append_newline (bool, optional): 
                Whether or not to append a newline character after
                the tree's string representation. 
                Defaults to False.
            outputlabel_mapper (Mapping[Node,str], optional): 
                A mapping function (`Node` to `str`) that maps a node
                to the label string that is to be written for it in 
                the output. 
                Defaults to `lambda n: n.get_label()` for regular 
                nodes (might differ for other kinds), so that a 
                `Node` is mapped to its `Node._label`, which is the 
                identification `label` (unique in its parent). 
                Please note that your custom implementation would 
                have to take the differences between the different 
                kinds of nodes into account.

        Returns:
            str: A string representation of this tree.
        """
        ret = []
        ret.append( \
            self._root.to_string(with_labels=with_labels,
                                 with_distances=with_distances,
                                 with_additional_info_nhx=with_additional_info_nhx,
                                 outputlabel_mapper=outputlabel_mapper))
        ret.append(';')
        if append_newline:
            ret.append(linesep)
        return ''.join(ret)
    


from newick.backend.path import Path
import pytest

def test_basic0():
    a = Path("R", [("A", 1.0)])
    assert len(a) == 2
    assert str(a) == "Path(R:0 -> A:1)"
    
def test_basic1():
    b = Path("R", [("B", 2.0), ("C", 2.2)])
    assert len(b) == 3
    assert str(b) == "Path(R:0 -> B:2 -> C:2.2)"

def test_basic3():
    c = Path("Root")
    assert len(c) == 1
    assert str(c) == "Path(Root:0)"
    
def test_getitem():
    b = Path("R", [("B", 2.0), ("C", 2.2)])
    assert b[1] == ("B", 2.0)
    with pytest.raises(Exception) as e:
        c = b[6]
    
def test_iter():
    b = Path("R", [("B", 2), ("C", 2.2), ("meep", 0)])
    for c in b:
        assert isinstance(c, tuple)
        assert len(c) == 2
        assert isinstance(c[0], str)
        assert isinstance(c[1], float)
        